use proc_macro2::{Span, TokenStream};
use quote::{format_ident, quote, quote_spanned};
use syn::{
    self, parse_quote, spanned::Spanned, AttributeArgs, FnArg, Ident, ItemTrait, NestedMeta,
    ReturnType, TraitItemMethod, Type,
};

use crate::utils::*;

pub fn expand(args: AttributeArgs, input: ItemTrait) -> TokenStream {
    let sync_proxy = create_proxy(&args, &input, false);
    let async_proxy = create_proxy(&args, &input, true);

    quote! {
        #sync_proxy

        #async_proxy
    }
}

pub fn create_proxy(args: &[NestedMeta], input: &ItemTrait, azync: bool) -> TokenStream {
    let mut iface_name = None;
    let mut default_path = None;
    let mut default_service = None;
    let mut has_introspect_method = false;

    let zbus = get_zbus_crate_ident();

    for arg in args {
        match arg {
            NestedMeta::Meta(syn::Meta::NameValue(nv)) => {
                if nv.path.is_ident("interface") || nv.path.is_ident("name") {
                    if let syn::Lit::Str(lit) = &nv.lit {
                        iface_name = Some(lit.value());
                    } else {
                        panic!("Invalid interface argument")
                    }
                } else if nv.path.is_ident("default_path") {
                    if let syn::Lit::Str(lit) = &nv.lit {
                        default_path = Some(lit.value());
                    } else {
                        panic!("Invalid path argument")
                    }
                } else if nv.path.is_ident("default_service") {
                    if let syn::Lit::Str(lit) = &nv.lit {
                        default_service = Some(lit.value());
                    } else {
                        panic!("Invalid service argument")
                    }
                } else {
                    panic!("Unsupported argument");
                }
            }
            _ => panic!("Unknown attribute"),
        }
    }

    let doc = get_doc_attrs(&input.attrs);
    let proxy_name = if azync {
        Ident::new(&format!("Async{}Proxy", input.ident), Span::call_site())
    } else {
        Ident::new(&format!("{}Proxy", input.ident), Span::call_site())
    };
    let ident = input.ident.to_string();
    let name = iface_name.unwrap_or(format!("org.freedesktop.{}", ident));
    let default_path = default_path.unwrap_or(format!("/org/freedesktop/{}", ident));
    let default_service = default_service.unwrap_or_else(|| name.clone());
    let mut methods = TokenStream::new();

    // FIXME: Create an enum type to hold these two.
    let (usage, wait) = if azync {
        (quote! { async }, quote! { .await })
    } else {
        (quote! {}, quote! {})
    };

    for i in input.items.iter() {
        if let syn::TraitItem::Method(m) = i {
            let method_name = m.sig.ident.to_string();
            if method_name == "introspect" {
                has_introspect_method = true;
            }

            let attrs = parse_item_attributes(&m.attrs, "dbus_proxy").unwrap();
            let is_property = attrs.iter().any(|x| x.is_property());
            let is_signal = attrs.iter().any(|x| x.is_signal());
            let has_inputs = m.sig.inputs.len() > 1;
            let name = attrs
                .iter()
                .find_map(|x| match x {
                    ItemAttribute::Name(n) => Some(n.to_string()),
                    _ => None,
                })
                .unwrap_or_else(|| {
                    pascal_case(if is_property && has_inputs {
                        assert!(method_name.starts_with("set_"));
                        &method_name[4..]
                    } else {
                        &method_name
                    })
                });
            let m = if is_property {
                gen_proxy_property(&name, &m, &usage, &wait)
            } else if is_signal {
                gen_proxy_signal(&name, &method_name, &m, &usage, &wait, azync)
            } else {
                gen_proxy_method_call(&name, &method_name, &m, &usage, &wait, azync)
            };
            methods.extend(m);
        }
    }

    if !has_introspect_method {
        methods.extend(quote! {
            #[doc = "Introspect the associated object, and return the XML description."]
            pub #usage fn introspect(&self) -> ::#zbus::fdo::Result<String> {
                self.0.introspect()#wait
            }
        });
    };

    let (proxy_doc, proxy_struct, connection) = if azync {
        let sync_proxy = Ident::new(&format!("{}Proxy", input.ident), Span::call_site());
        let doc = format!("Asynchronous sibling of [`{}`].", sync_proxy);
        let connection = quote! { ::#zbus::azync::Connection };
        let proxy = quote! { ::#zbus::azync::Proxy };

        (doc, proxy, connection)
    } else {
        let doc = String::from("");
        let connection = quote! { ::#zbus::Connection };
        let proxy = quote! { ::#zbus::Proxy };

        (doc, proxy, connection)
    };

    quote! {
        #[doc = #proxy_doc]
        #(#doc)*
        #[derive(Debug)]
        pub struct #proxy_name<'c>(#proxy_struct<'c>);

        impl<'c> #proxy_name<'c> {
            /// Creates a new proxy with the default service & path.
            pub fn new(conn: &#connection) -> ::#zbus::Result<Self> {
                Ok(Self(#proxy_struct::new(
                    conn,
                    #default_service,
                    #default_path,
                    #name,
                )?))
            }

            /// Creates a new proxy with the default service & path, taking ownership of `conn`.
            pub fn new_owned(conn: #connection) -> ::#zbus::Result<Self> {
                Ok(Self(#proxy_struct::new_owned(
                    conn,
                    String::from(#default_service),
                    String::from(#default_path),
                    String::from(#name),
                )?))
            }

            /// Creates a new proxy for the given `destination` and `path`.
            pub fn new_for<E>(
                conn: &#connection,
                destination: &'c str,
                path: impl std::convert::TryInto<#zbus::export::zvariant::ObjectPath<'c>, Error = E>,
            ) -> ::#zbus::Result<Self>
            where
                ::#zbus::Error: From<E>,
            {
                Ok(Self(#proxy_struct::new(
                    conn,
                    destination,
                    path,
                    #name,
                )?))
            }

            /// Same as `new_for` but takes ownership of the passed arguments.
            pub fn new_for_owned<E>(
                conn: #connection,
                destination: String,
                path: impl std::convert::TryInto<#zbus::export::zvariant::ObjectPath<'static>, Error = E>,
            ) -> ::#zbus::Result<Self>
            where
                ::#zbus::Error: From<E>,
            {
                Ok(Self(#proxy_struct::new_owned(
                    conn,
                    destination,
                    path,
                    #name.to_owned(),
                )?))
            }

            /// Creates a new proxy for the given `path`.
            pub fn new_for_path<E>(
                conn: &#connection,
                path: impl std::convert::TryInto<#zbus::export::zvariant::ObjectPath<'c>, Error = E>,
            ) -> ::#zbus::Result<Self>
            where
                ::#zbus::Error: From<E>,
            {
                Ok(Self(#proxy_struct::new(
                    conn,
                    #default_service,
                    path,
                    #name,
                )?))
            }

            /// Same as `new_for_path` but takes ownership of the passed arguments.
            pub fn new_for_owned_path<E>(
                conn: #connection,
                path: impl std::convert::TryInto<#zbus::export::zvariant::ObjectPath<'static>, Error = E>,
            ) -> ::#zbus::Result<Self>
            where
                ::#zbus::Error: From<E>,
            {
                Ok(Self(#proxy_struct::new_owned(
                    conn,
                    #default_service.to_owned(),
                    path,
                    #name.to_owned(),
                )?))
            }

            /// Consumes `self`, returning the underlying `zbus::Proxy`.
            pub fn into_inner(self) -> #proxy_struct<'c> {
                self.0
            }

            /// The reference to the underlying `zbus::Proxy`.
            pub fn inner(&self) -> &#proxy_struct {
                &self.0
            }

            #methods
        }

        impl<'c> std::ops::Deref for #proxy_name<'c> {
            type Target = #proxy_struct<'c>;

            fn deref(&self) -> &Self::Target {
                &self.0
            }
        }

        impl<'c> std::ops::DerefMut for #proxy_name<'c> {
            fn deref_mut(&mut self) -> &mut Self::Target {
                &mut self.0
            }
        }

        impl<'c> std::convert::AsRef<#proxy_struct<'c>> for #proxy_name<'c> {
            fn as_ref(&self) -> &#proxy_struct<'c> {
                &*self
            }
        }

        impl<'c> std::convert::AsMut<#proxy_struct<'c>> for #proxy_name<'c> {
            fn as_mut(&mut self) -> &mut #proxy_struct<'c> {
                &mut *self
            }
        }

        impl<'c> #zbus::export::zvariant::Type for #proxy_name<'c> {
            fn signature() -> #zbus::export::zvariant::Signature<'static> {
                #zbus::export::zvariant::OwnedObjectPath::signature()
            }
        }

        impl<'c> #zbus::export::serde::ser::Serialize for #proxy_name<'c> {
            fn serialize<S>(&self, serializer: S) -> std::result::Result<S::Ok, S::Error>
            where
                S: #zbus::export::serde::ser::Serializer,
            {
                String::serialize(&self.inner().path().to_string(), serializer)
            }
        }
    }
}

fn gen_proxy_method_call(
    method_name: &str,
    snake_case_name: &str,
    m: &TraitItemMethod,
    usage: &TokenStream,
    wait: &TokenStream,
    azync: bool,
) -> TokenStream {
    let zbus = get_zbus_crate_ident();
    let doc = get_doc_attrs(&m.attrs);
    let args = m.sig.inputs.iter().filter_map(|arg| arg_ident(arg));
    let attrs = parse_item_attributes(&m.attrs, "dbus_proxy").unwrap();
    let proxy_object = attrs.iter().find_map(|x| match x {
        ItemAttribute::Object(o) => {
            if azync {
                Some(format!("Async{}Proxy", o))
            } else {
                Some(format!("{}Proxy", o))
            }
        }
        _ => None,
    });
    if let Some(proxy_name) = proxy_object {
        let method = Ident::new(snake_case_name, Span::call_site());
        let proxy = Ident::new(&proxy_name, Span::call_site());
        let inputs = &m.sig.inputs;
        let (_, ty_generics, where_clause) = m.sig.generics.split_for_impl();
        let signature = if where_clause.is_some() {
            quote! {
                fn #method#ty_generics(#inputs) -> ::#zbus::Result<#proxy<'_>>
                #where_clause,
            }
        } else {
            quote! {
                fn #method(#inputs) -> ::#zbus::Result<#proxy<'_>>
            }
        };

        quote! {
            #(#doc)*
            pub #usage #signature {
                let object_path: ::#zbus::export::zvariant::OwnedObjectPath =
                    self.0.call(
                        #method_name,
                        &(#(#args),*),
                    )
                    #wait?;
                let proxy = #proxy::new_for_owned_path(
                    self.0.connection().clone(),
                    object_path.into_inner(),
                )?;
                Ok(proxy)
            }
        }
    } else {
        let signature = &m.sig;

        quote! {
            #(#doc)*
            pub #usage #signature {
                let reply = self.0.call(#method_name, &(#(#args),*))#wait?;
                Ok(reply)
            }
        }
    }
}

fn gen_proxy_property(
    property_name: &str,
    m: &TraitItemMethod,
    usage: &TokenStream,
    wait: &TokenStream,
) -> TokenStream {
    let doc = get_doc_attrs(&m.attrs);
    let signature = &m.sig;
    if signature.inputs.len() > 1 {
        let value = arg_ident(signature.inputs.last().unwrap()).unwrap();
        quote! {
            #(#doc)*
            pub #usage #signature {
                Ok(self.0.set_property(#property_name, #value)#wait?)
            }
        }
    } else {
        // This should fail to compile only if the return type is wrong,
        // so use that as the span.
        let body_span = if let ReturnType::Type(_, ty) = &signature.output {
            ty.span()
        } else {
            signature.span()
        };
        let body = quote_spanned! {body_span =>
            Ok(self.0.get_property::<_>(#property_name)#wait?)
        };
        quote! {
            #(#doc)*
            pub #usage #signature {
                #body
            }
        }
    }
}

fn gen_proxy_signal(
    signal_name: &str,
    snake_case_name: &str,
    m: &TraitItemMethod,
    usage: &TokenStream,
    wait: &TokenStream,
    azync: bool,
) -> TokenStream {
    let zbus = get_zbus_crate_ident();
    let doc = get_doc_attrs(&m.attrs);
    let method = format_ident!("connect_{}", snake_case_name);
    let input_types: Vec<Box<Type>> = m
        .sig
        .inputs
        .iter()
        .filter_map(|arg| match arg {
            FnArg::Typed(p) => Some(p.ty.clone()),
            _ => None,
        })
        .collect();
    let args: Vec<Ident> = m
        .sig
        .inputs
        .iter()
        .filter_map(|arg| arg_ident(arg).cloned())
        .collect();
    let handler = if azync {
        quote! { FnMut(#(#input_types),*) -> ::#zbus::export::futures_core::future::BoxFuture<'static, ::#zbus::Result<()>> }
    } else {
        quote! { FnMut(#(#input_types),*) -> ::#zbus::Result<()> }
    };

    let (proxy_method, link) = if azync {
        (
            "zbus::azync::Proxy::connect_signal",
            "https://docs.rs/zbus/latest/zbus/azync/struct.Proxy.html#method.connect_signal",
        )
    } else {
        (
            "zbus::Proxy::connect_signal",
            "https://docs.rs/zbus/latest/zbus/struct.Proxy.html#method.connect_signal",
        )
    };
    let gen_doc = format!(
        " Connect the handler for the `{}` signal. This is a convenient wrapper around [`{}`]({}).",
        signal_name, proxy_method, link,
    );

    let mut generics = m.sig.generics.clone();
    generics.params.push(parse_quote!(__H));
    {
        let where_clause = generics.where_clause.get_or_insert(parse_quote!(where));
        where_clause
            .predicates
            .push(parse_quote!(__H: #handler + Send + 'static));
    }

    let (_, ty_generics, where_clause) = generics.split_for_impl();
    quote! {
        #[doc = #gen_doc]
        #(#doc)*
        pub #usage fn #method#ty_generics(
            &self,
            mut handler: __H,
        ) -> ::#zbus::fdo::Result<::#zbus::SignalHandlerId>
        #where_clause,
        {
            self.0.connect_signal(#signal_name, move |m| {
                let (#(#args),*) = m.body().expect("Incorrect signal signature");

                handler(#(#args),*)
            })#wait
        }
    }
}
